#!/usr/bin/env python3

import argparse
import os
import subprocess
import sys
import time
import tempfile
from pathlib import Path

import watchdog.events
import watchdog.observers


docker_script = """
FROM debian:testing
RUN apt-get update && apt-get install -y asciidoctor
RUN gem install asciidoctor-diagram
RUN gem install asciidoctor-pdf
"""


def render_doctype(config):
    print(f"Rendering {config.command} '{config.doc_path}'...", end="")
    sys.stdout.flush()

    doc_path = Path(config.doc_path).absolute()

    if config.not_containerized:
        docker_args = []

        containerized_src = doc_path.parent.absolute()
        containerized_dst = Path(config.dst if config.dst is not None else doc_path.parent).absolute()

        doc_prefix = ""
    else:
        containerized_src = "/src"
        containerized_dst = "/dst"

        abs_src_dir_path = doc_path.parent.absolute()
        abs_dst_dir_path = Path(config.dst).absolute() if config.dst is not None else abs_src_dir_path

        docker_args = [
            "docker",
            "run",
            "-v", f"{abs_src_dir_path}:{containerized_src}",
            "-v", f"{abs_dst_dir_path}:{containerized_dst}",
            "golfdoc",
        ]

        doc_prefix = "." + os.path.commonprefix([containerized_src, doc_path])
        
    if config.dst is not None and not os.path.exists(config.dst):
        os.makedirs(config.dst)

    containerized_doc_path = os.path.abspath(os.path.join(containerized_src, doc_prefix, os.path.basename(config.doc_path)))

    args = ([] if config.not_containerized else docker_args) + [
        "asciidoctor",
        "-S", config.security_level,
        "--backend", config.backend,
        "--failure-level=WARN",
        "-a attribute-missing=warn",
        "--require", "asciidoctor-diagram",
        "--require", "asciidoctor-pdf",
        "--doctype", config.command,
        "--destination-dir", containerized_dst,
        "--source-dir", containerized_src,
        containerized_doc_path
    ]

    completed_proc = subprocess.run(args)

    ok = completed_proc.returncode == 0

    print("done." if ok else "FAILED.")

    return ok


doctype_handlers = {
    'book': render_doctype,
    'article': render_doctype,
}


def add_render_args(argparser):
    argparser.add_argument("-S", "--security-level", default="secure", choices="secure,server,safe,unsafe".split(","))
    argparser.add_argument("-w", "--watch", action="store_true", default=False)
    argparser.add_argument("-s", "--src")
    argparser.add_argument("-b", "--backend", choices="pdf,html5", default="html5")
    argparser.add_argument("-d", "--dst")
    argparser.add_argument("-N", "--not-containerized", action="store_true", default=False)
    argparser.add_argument("doc_path")


def build_config():
    argparser = argparse.ArgumentParser()

    subparsers = argparser.add_subparsers(dest='command', help='command help')

    container_parser = subparsers.add_parser('container')
    container_parser.add_argument('subcommand', choices=['build'])

    for doctype in doctype_handlers.keys():
        doctype_parser = subparsers.add_parser(doctype)
        add_render_args(doctype_parser)

    config = argparser.parse_args(sys.argv[1:])

    return config


def fatal_error(msg):
    print("ERROR: " + msg, file=sys.stderr)
    return 20


def build_container():
    with tempfile.TemporaryDirectory(prefix="golfdoc") as build_dir:
        with open(os.path.join(build_dir, "Dockerfile"), "w") as dockerfile:
            dockerfile.write(docker_script)

        args = [
            "docker",
            "build",
            "-t", "golfdoc",
            "."
        ]

        subprocess.run(args, cwd=build_dir)


def watch(config, handler):
    patterns = ["*"]
    ignore_patterns = [".git", ".gitignore"]
    ignore_directories = False
    case_sensitive = True
    recurse = True

    last_result = 20

    def handle_with_result(event):
        nonlocal config
        nonlocal handler
        nonlocal last_result

        last_result = handler(config)

    fs_change_handler = watchdog.events.PatternMatchingEventHandler(patterns, ignore_patterns, ignore_directories, case_sensitive)
    fs_change_handler.on_created = handle_with_result
    fs_change_handler.on_deleted = handle_with_result
    fs_change_handler.on_modified = handle_with_result
    fs_change_handler.on_moved = handle_with_result

    observer = watchdog.observers.Observer()
    observer.schedule(fs_change_handler, config.src, recursive=recurse)

    observer.start()
    try:
        print(f"Monitoring {config.src} for changes and re-rendering.  Send Ctrl-C to quit.")

        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
        observer.join()

    return 


def dispatch_to_renderer(config):
    try:
        handler = doctype_handlers[config.command]
    except IndexError:
        return fatal_error(f"No such doc_type: {config.command}")

    ok = handler(config)

    if config.watch:
        ok = watch(config, handler) 

    return ok


def main():
    config = build_config()

    if config.command == "container":
        ok = build_container()
    elif config.command is None:
        return fatal_error("Need a command.  Try --help, or read the docs!")
        
    else:
        ok = dispatch_to_renderer(config)

    return 0 if ok else 5

